@@ -3,10 +3,15 @@ package Search::ContextGraph;
 use strict;
 use warnings;
 use Carp;
+use DynaLoader;
+
 use base "Storable";
 
 
-our $VERSION = '0.05';
+our $VERSION = '0.06';
+our @ISA = qw/DynaLoader Storable/;
+
+bootstrap Search::ContextGraph $VERSION;
 
 
 =head1 NAME
@@ -73,28 +78,73 @@ posed by latent semantic indexing (LSI).
 
 =item new %PARAMS
 
-Object constructor.  
+Object constructor.   Possible parameters:
+
+=over
+
+=item * debug LEVEL
+
+Set this to 1 or 2 to turn on verbose debugging output
+
+=item * xs
+
+When true, tells the module to use compiled C internals.  This reduces
+memory requirements by about 60%, but actually runs a little slower than the 
+pure Perl version.  Don't bother to turn it on unless you have a huge graph. 
+Default is pure Perl.
+
+BUG: using the compiled version makes it impossible to store the graph to disk.
+
+=item * START_ENERGY
+
+Initial energy to assign to a query node.  Default is 100.
+
+=item * ACTIVATE_THRESHOLD
+
+Minimal energy needed to propagate search along the graph.  Default is 1.
+
+=item * COLLECT_THRESHOLD
+
+Minimal energy needed for a node to enter the result set.  Default is 1.
+
+=back 
 
 =cut
 
+
 sub new {
 	my ( $class, %params) = @_;
-	bless 
+	my $obj = bless 
 		{ debug => 0,
+		  fast => 0,
 		  START_ENERGY => 100,
 		  ACTIVATE_THRESHOLD => 1,
-		  COLLECT_THRESHOLD => 1,
+		  COLLECT_THRESHOLD => .2,
 	      %params,
 	      neighbors => {}, 
 
 		}, 
 	$class;
+	if ( $obj->{'xs'} ) {
+		my $graph = Search::ContextGraph::Graph->new(
+			$obj->{START_ENERGY},
+			$obj->{ACTIVATE_THRESHOLD},
+			$obj->{COLLECT_THRESHOLD},
+		);
+		$obj->{Graph} = $graph;
+		$obj->{'next_free_id'} = 0;
+		$obj->{'node_map'} = {};
+		
+	}
+	return $obj;
+		
 }
 
 =item [get|set]_activate_threshold
 
 Accessor for node activation threshold value.  This value determines how far 
-energy can spread in the graph
+energy can spread in the graph.  Lower it to increase the number of results.
+Default is 1.
 
 =cut
 
@@ -105,18 +155,48 @@ sub set_activate_threshold {	$_[0]->{'ACTIVATE_THRESHOLD'} = $_[1] }
 =item [get|set]_collect_threshold
 
 Accessor for collection threshold value.  This determines how much energy a
-node must have to make it into the result set.
+node must have to make it into the result set.  Lower it to increase the 
+number of results.   Default is 1.
 
 =cut
 
-sub get_collect_threshold {  $_[0]->{'COLLECT_THRESHOLD'} }
-sub set_collect_threshold {	 $_[0]->{'COLLECT_THRESHOLD'} = $_[1] }
+sub get_collect_threshold {  
+	return ( $_[0]->{'xs'} ? 
+		$_[0]->{Graph}->collectionThreshold :
+		$_[0]->{'COLLECT_THRESHOLD'})
+ }
+ 
+sub set_collect_threshold {	
+	 my ( $self, $newval ) = @_;
+	 
+	 return if $newval =~ /[^\d.]/;
+	 if ( $newval < 0 ) {
+	 	$newval = 0;
+	 }
+	 
+	 
+	 $self->{Graph}->collectionThreshold( $newval )
+	 	if $self->{'xs'};
+	 
+	 $self->{'COLLECT_THRESHOLD'} = $newval;
+}
 
+=item set_debug_mode [012]
+
+Turns debugging on or off.  1 is verbose, 2 is very verbose, 0 is off.
+
+=cut
+
+sub set_debug_mode {
+	my ( $self, $mode ) = @_;
+	$self->{'debug'} = $mode;
+}
 
 =item [get|set]_initial_energy
 
-Accessor for initial energy value at the query node.  The higher this value,
-the larger the result set
+Accessor for initial energy value at the query node.  This controls how 
+much energy gets poured into the graph at the start of the search.
+Increase this value to get more results from your queries.
 
 =cut
 
@@ -127,10 +207,8 @@ sub set_initial_energy { $_[0]->{'START_ENERGY'} = $_[1] }
 =item load_from_tdm TDM_FILE [, LM_FILE]
 
 Opens and loads a term-document matrix (TDM) file to initialize the graph.
-Optionally also opens and loads a document link matrix (DLM) file of document-to-document
-links.  The TDM encodes information about term-to-document links, while the DLM
-file holds information about inter-document links, like hyperlinks or citation data.
-For notes on these file formats, see the README file
+The TDM encodes information about term-to-document links.
+For notes on the proper file format, see the README file
 Note that document-document links are NOT YET IMPLEMENTED.
 
 =cut
@@ -150,30 +228,57 @@ sub load_from_tdm {
 
 Given a list of nodes, returns a hash of nearest nodes with relevance values,
 in the format NODE => RELEVANCE, for all nodes above the threshold value. 
-(You probably want one of search, find_similar, or mixed_search instead).
+(You probably want one of L<search>, L<find_similar>, or L<mixed_search> instead).
 
 =cut
 
 sub raw_search {
 	my ( $self, @query ) = @_;
 
-	$self->_clear();
-	foreach ( @query ) {
-		$self->_energize( $_, $self->{'START_ENERGY'} );
-	}
-	my $results_ref = $self->_collect();
+	
+	my $results_ref;
+	
+	### XS VERSION  #######
+	if ( $self->{'xs'} ) {
+		
+		$self->{'Graph'}->reset_graph();
+		my $map = $self->{'node_map'};
 
+		foreach my $node ( map { $map->{$_} } @query ) {
+
+			$self->{'Graph'}->energize_node( $node, $self->{'START_ENERGY'}, 1);
+		}
+		my $results_arref = $self->{Graph}->collect_results();
+
+		my %hash;
+
+		foreach my $res ( @{$results_arref} ) {
+			my $key =  $self->{'id_map'}[$res->[0]] || next;
+			$hash{$key} = $res->[1] || 0;
+		}
+		$results_ref = \%hash;
+		
+				
+	#### PURE PERL VERSION #####
+	} else {
+	    $self->_clear();
+		foreach ( @query ) {
+			$self->_energize( $_, $self->{'START_ENERGY'});
+		}
+		$results_ref = $self->_collect();
+	}	
+	 
 	return $results_ref;
 }
 
-=item set_debug_mode
+=item debug_on, debug_off
 
-Turns verbose comments on if given a true value as its argument
+Toggles debug mode
 
 =cut
 
-sub set_debug_mode { $_[0]->{'debug'} = $_[1] }
-
+sub debug_on { $_[0]->{'debug'} = 1 }
+sub debug_off { $_[0]->{'debug'} = 0 }
 
 
 # Opens and reads a term-document matrix (TDM) file.  The format for this file
@@ -182,39 +287,97 @@ sub set_debug_mode { $_[0]->{'debug'} = $_[1] }
 
 sub _read_tdm {
 	my ( $self, $file ) = @_;
-	print "Loading TDM...\n" if $self->{'debug'};
-
+	print "Loading TDM...\n" if $self->{'debug'} > 1;
 
+	croak "File does not exist" unless -f $file;
 	open my $fh, $file or croak "Could not open $file: $!";
 	for ( 1..4 ){
 		my $skip = <$fh>;
 	}
 	my %neighbors;
-	my $doc = 0;
-	while (<$fh>) {
-		chomp;
-		my ( $count, %vals ) = split;
-		while ( my ( $n, $v ) = each %vals ) {
-			$neighbors{"D:$doc"}{"T:$n"} = $v;
-			$neighbors{"T:$n"}{"D:$doc"} = $v;
+	my $doc = 0;	
+	
+	
+	######### XS VERSION ##############
+	if ( $self->{'xs'} ) {
+		
+		my $map = $self->{'node_map'}; # shortcut alias
+		while (<$fh>) {
+			chomp;
+			my $dindex = $self->_add_node( "D:$doc", 2 );
+			#warn "Added node $doc\n";
+			my ( $count, %vals ) = split;
+			while ( my ( $term, $edge ) = each %vals ) {
+				$self->{'term_count'}{$term}++;
+				my $tnode = "T:$term";
+				
+				my $tindex = ( defined $map->{$tnode} ?
+								$map->{$tnode} : 
+							 	$self->_add_node( $tnode, 1 )
+							);
+				$self->{Graph}->set_edge( $dindex, $tindex, $edge );				
+			}
+			$doc++;
 		}
-		$doc++;
+		
+	####### PURE PERL VERSION ##########
+	} else {
+		while (<$fh>) {
+			chomp;
+			my $dnode = "D:$doc";
+			my ( $count, %vals ) = split;
+			while ( my ( $term, $edge ) = each %vals ) {
+				$self->{'term_count'}{$term}++;
+				my $tnode = "T:$term";
+				
+				$neighbors{$dnode}{$tnode} = $edge;
+				$neighbors{$tnode}{$dnode} = $edge;
+			}
+			$doc++;
+		}
+		$self->{'neighbors'} = \%neighbors;	
 	}
-	$self->{'neighbors'} = \%neighbors;
-	print "Loaded.\n" if $self->{'debug'};
+	
+	print "Loaded.\n" if $self->{'debug'} > 1;
 	$self->{'from_TDM'} = 1;
 	$self->{'doc_count'} = $doc;
 }
 
 
+
+# XS version only
+#
+# This sub maintains a mapping between node names and integer index
+# values. 
+
+sub _add_node {
+	my ( $self, $node_name, $type ) = @_;
+	croak "Must provide a type" unless $type;
+	croak "Must provide a node name" unless $node_name;
+	croak "This node already exists" if 
+		 $self->{'node_map'}{$node_name};
+		
+	my $new_id = $self->{'next_free_id'}++;
+	$self->{'node_map'}{$node_name} = $new_id;
+	$self->{'id_map'}[$new_id] = $node_name;
+	$self->{'Graph'}->add_node( $new_id, $type );
+	
+	return $new_id;
+}
+
 =item add_documents %DOCS
 
 Load up the search engine with documents in the form
-TITLE => WORDHASH, where WORDHASH is a reference to a hash of terms
-and occurence counts.  In other words,
+TITLE => WORDS, where WORDS is either a  reference to a hash of terms
+and occurence counts, or a reference to an array of words.
+For example:
 
 	TITLE => { WORD1 => COUNT1, WORD2 => COUNT2 ... }
 
+or
+
+	TITLE => [ WORD1, WORD2, WORD3 ]
+
 
 =cut
 
@@ -223,13 +386,27 @@ sub add_documents {
 	my ( $self, %incoming_docs ) = @_;
 
 	my @doc_list = keys %incoming_docs; # Make sure these remain in same order
-
+	
 	# Add word and document nodes to the graph
 	foreach my $doc ( @doc_list ) {
 		my $dnode = 'D:'.$doc;
 		croak "Tried to add document with duplicate title: \"$doc\"\n"
-			if exists( $self->{'neighbors'}{$dnode} );
-
+			if exists( $self->{'neighbors'}{$dnode} ) or 
+			exists $self->{'node_map'}{$dnode};
+			
+		croak "Not a reference" unless ref( $incoming_docs{$doc});
+		my $type = ref( $incoming_docs{$doc});
+		my @list;
+		if ( $type eq 'HASH' ) {
+			@list = keys %{ $incoming_docs{$doc} }
+		} elsif ( $type eq 'ARRAY' ) {
+			@list = @{$incoming_docs{$doc}};
+		} else {
+			croak "Wrong type of reference"
+		}
+		foreach my $term ( @list  ){
+			$self->{'term_count'}{$term}++;
+		}
 	}
 
 
@@ -241,22 +418,30 @@ sub add_documents {
 		# Every word in this document
 		my @edges;
 		my $dnode = 'D:'.$doc;
-
-		foreach my $term ( keys %{ $incoming_docs{$doc} } ) {		
+		my $type = ref( $incoming_docs{$doc} ); 
+		my @list = ( $type eq 'HASH' ?
+					 keys %{ $incoming_docs{$doc} } :
+					 @{ $incoming_docs{$doc} }
+					);
+					
+		foreach my $term ( @list ) {		
 
 			my $tnode = 'T:'.$term;
 
-			my $lcount = $incoming_docs{$doc}{$term};
+			my $lcount = ( $type eq 'HASH' ?
+						   $incoming_docs{$doc}{$term} :
+						   1);
+						   
 			my $l_weight = log( $lcount ) + 1;
 
 			# We calculate global weight in the loop to conserve memory
-			my $global_count = $self->degree( $tnode ) + 1;
+			my $global_count = $self->{'term_count'}{$term};
 			my $g_weight = log( $self->{'doc_count'} /  $global_count ) + 1;
 
 			# edge weight is local weight * global weight
 			my $t_weight = ( $g_weight * $l_weight );
-			print "Calculated edge $dnode -> $tnode with weight $t_weight", "\n"
-				 if $self->{debug};
+			#print "Calculated edge $dnode -> $tnode with weight $t_weight", "\n"
+				# if $self->{debug};
 			push @edges, [ $dnode, $tnode, $t_weight ];
 		}
 
@@ -266,13 +451,34 @@ sub add_documents {
 		$sum += $_->[2] foreach @edges;
 		$_->[2]/= $sum foreach @edges;
 
-		foreach my $e ( @edges ) {
-			print "final weight is ", $e->[2], "\n" if $self->{debug};
-			$self->set_edge( @{$e} );
+		
+		#### XS VERSION ###############
+		if ( $self->{'xs'} ) {
+			my $map = $self->{'node_map'};
+			foreach my $e ( @edges ) {
+				my ( $doc, $term, $edge ) = @{$e};
+				my $dindex = ( exists $map->{$doc}  
+								? $map->{$doc}  
+								: $self->_add_node( $doc, 2 ) );
+								
+				my $tindex =  ( exists $map->{$term} 
+								? $map->{$term}
+								: $self->_add_node( $term, 1 ));
+								
+				$self->{Graph}->set_edge( $dindex, $tindex, $edge );
+			}
+		
+		#### PURE PERL VERSION ########
+		} else {
+			foreach my $e ( @edges ) {
+				my ( $doc, $term, $edge ) = @{$e};
+				print "$doc $term edge weight is ", $edge, "\n" if $self->{debug};
+				$self->set_edge( $doc, $term, $edge );
+			}
 		}
+		
 	}
 
-
 }
 
 
@@ -293,18 +499,14 @@ sub retrieve {
 sub dump_words {
 
 	my ( $self ) = @_;
-	my %words;
-	foreach my $n ( keys %{ $self->{'neighbors'} } ) {
-		$words{$n}++ if $n =~ s/^T://o;
-	}
-
-	return  keys %words;
+	return keys %{ $self->{'term_count'} };
 }
 
 =item search @QUERY
 
-Searches the graph for all of the words in @QUERY.   No support yet for 
-document similarity searches, but it's coming.  Returns a pair of hashrefs:
+Searches the graph for all of the words in @QUERY.  Use find_similar if you
+want to do a document similarity instead, or mixed_search if you want
+to search on any combination of words and documents.  Returns a pair of hashrefs:
 the first a reference to a hash of docs and relevance values, the second to 
 a hash of words and relevance values.
 
@@ -321,9 +523,9 @@ sub search {
 
 =item find_similar @DOCS
 
-Given an array of document identifiers, returns a pair of hashrefs.
-First hashref is to a hash of docs and relevance values, second
-is to a hash of words and relevance values.
+Given an array of document identifiers, performs a similarity search 
+and  returns a pair of hashrefs. First hashref is to a hash of docs and relevance
+ values, second is to a hash of words and relevance values.
 
 =cut
 
@@ -359,6 +561,7 @@ sub mixed_search {
 	my $tref = $incoming->{'terms'} || [];
 	my $dref = $incoming->{'docs'}  || [];
 
+	
 	my @dnodes = $self->_nodeify( 'D', @{$dref} );
 	my @tnodes = $self->_nodeify( 'T', @{$tref} );
 
@@ -368,6 +571,26 @@ sub mixed_search {
 }
 
 
+=item store FILENAME
+
+Stores the object to a file for later use.  Not compatible (yet)
+with compiled XS version, which will give a fatal error.
+
+=cut
+
+sub store {
+	my ( $self, @args ) = @_;
+	if ( $self->{'xs'} ) {
+		croak "Cannot store object when running in XS mode.";
+	} else {
+		$self->SUPER::store(@args);
+	}
+}
+
+
+# Partition - internal method.
+# Takes a result set and splits it into two hashrefs - one for
+# words and one for documents
 
 sub _partition {
 	my ( $e ) = @_;
@@ -389,12 +612,15 @@ sub _nodeify {
 	foreach my $item ( @list ) {
 		my $name = $prefix.':'.$item;
 		croak "Node $name not found"
-			unless defined $self->{'neighbors'}{$name};
+			unless $self->{'xs'} 
+			or defined $self->{'neighbors'}{$name};
 		push @nodes, $name;
 	}
 	return @nodes;
 }
 
+
+
 sub degree {
 	my ($self, $node ) = @_;
 	if ( exists ( $self->{'neighbors'}{$node} ) ) {
@@ -403,6 +629,7 @@ sub degree {
 }
 
 
+
 sub set_edge {
 	my ( $self, $source, $sink, $value ) = @_;
 	croak "No source node" unless defined $source;
@@ -412,6 +639,12 @@ sub set_edge {
 	if ( $value > 1 ) {
 		croak "found edge exceeding unit weight\n";
 	}
+	
+	if ( defined $self->{'neighbors'}{$source}{$sink} or
+		 defined $self->{'neighbors'}{$sink}{$source} ){
+		 	die "Found existing edge ( $sink, $source )";
+		 }
+		
 	$self->{'neighbors'}{$source}{$sink} = $value;
 	$self->{'neighbors'}{$sink}{$source} = $value;
 }
@@ -428,7 +661,13 @@ sub _clear {
 
 sub _collect {
 	my ( $self ) = @_;
-	return $self->{'energy'};
+	my $e = $self->{'energy'};
+	my $result = {};
+	foreach my $k ( keys %{$self->{'energy'}} ) {
+		next unless $e->{$k} > $self->{'COLLECT_THRESHOLD'};
+		$result->{$k} = $e->{$k};
+	}
+	return $result;
 }
 
  #  Assign a starting energy ENERGY to NODE, and recursively distribute  the 
@@ -441,21 +680,18 @@ sub _energize {
 
 
 	return unless defined $self->{'neighbors'}{$node};
-
+	my $orig = $self->{'energy'}{$node} || 0;
 	$self->{'energy'}->{$node} += $energy;
-	$self->{'indent'}++;
+	$self->{'depth'}++;
 
 	if ( $self->{'debug'} > 1 ) {
-		print ' ' x $self->{'indent'};
-		print "Energizing node $node with energy $energy\n";
+		print '   ' x $self->{'depth'};
+		print "$node: energizing  $orig + $energy\n";
 	}
 	#sleep 1;
 	my $degree = scalar keys %{ $self->{'neighbors'}->{$node} };
 
-	if ( defined $self->{'debug'} && $self->{'debug'} > 1) {
-		print  ' ' x $self->{'indent'};
-		print "Node $node has $degree neighbors\n" if $self->{'debug'}	
-	}
+
 
 	croak "Fatal error: encountered node of degree zero" unless $degree;
 	my $subenergy = $energy / $degree;
@@ -468,33 +704,26 @@ sub _energize {
 	if ( $degree == 1 and  $energy < $self->{'START_ENERGY'} ) {
 
 		#do nothing
-
-
-	# If the search starts on a singleton, we want to make sure the energy
-	# gets passed along to the neighbor node, no matter how puny the edge weight
-	# connecting the singleton term to its neighbor.  We diminish it a little bit,
-	# just in case the neighbor is also a degree 1 node - otherwise the
-	# energy would  bounce back and forth forever
-
-	} elsif ( $degree == 1 ) {
-		my ($neighbor) = keys %{ $self->{'neighbors'}{$node} };
-		$self->_energize( $neighbor, $self->{'START_ENERGY'} * .9 );
-
-
-	# 
+	
 	} elsif ( $subenergy > $self->{ACTIVATE_THRESHOLD} ) {
-		print "\tsubenergy is $subenergy\n" if $self->{'debug'} > 1;
+		print '   ' x $self->{'depth'}, 
+		"$node: propagating subenergy $subenergy to $degree neighbors\n"
+		 if $self->{'debug'} > 1;
 		foreach my $neighbor ( keys %{ $self->{'neighbors'}{$node} } ) {
 			my $edge = $self->{'neighbors'}{$node}{$neighbor};
 			my $weighted_energy = $subenergy * $edge;
+			print '   ' x $self->{'depth'}, 
+			" edge $edge ($node, $neighbor)\n"
+		 		if $self->{'debug'} > 1;
 			$self->_energize( $neighbor, $weighted_energy );
 		} 
 	}	
-	$self->{'indent'}--;	
+	$self->{'depth'}--;	
 	return 1;
 }
 
 
+
 sub get_neighbors {
 	my ( $self, $node ) = @_;
 	my @list;
@@ -504,6 +733,10 @@ sub get_neighbors {
 	return @list;
 }
 
+sub DESTROY {
+	undef $_[0]->{Graph}
+}
+
 1;
 
 =back
@@ -512,11 +745,13 @@ sub get_neighbors {
 
 =over
 
-=item Document-document links are not yet implemented
+=item * Document-document links are not yet implemented
 
-=item No way to delete nodes once they're in the graph
+=item * No way to delete nodes once they're in the graph
 
-=item No way to break edges once they're in the graph
+=item * No way to break edges once they're in the graph
+
+=item * Can't store graph if using compiled C internals
 
 =back
 
@@ -527,17 +762,24 @@ Maciej Ceglowski E<lt>maciej@ceglowski.comE<gt>
 The technique used here was developed in 2003 by John Cuadrado, and later
 found to have antecedents in the spreading activation approach described 
 in a 1981 doctoral dissertation by Scott Preece.
+XS implementation thanks to Schuyler Erle.
 
 =head1 CONTRIBUTORS 
-
-Ken Williams
-Leon Brocard
+	
+	Schuyler Erle
+	Ken Williams
+	Leon Brocard  
 
 =head1 COPYRIGHT AND LICENSE
 
+Perl module:
 (C) 2003 Maciej Ceglowski
 
-This program may be distributed under the same terms as Perl itself.
+XS Implementation:
+(C) 2003 Maciej Ceglowski, Schuyler Erle
+
+This program is free software, distributed under the GNU Public License.
+See LICENSE for details.
 
 
 =cut
